import os
import torch
from typing import List
from torch.utils.data.dataset import Dataset
from pycocotools.coco import COCO
from nas_ood_single.search.models.utils.augmentations import *

coco_ids = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

coco_names = ["bike", "bus", "car", "motor", "person", "rider", "traffic light", "traffic sign", "train", "truck"]

colors = [(67, 68, 113), (130, 45, 169), (2, 202, 130), (127, 111, 90), (92, 136, 113),
          (33, 250, 7), (238, 92, 104), (0, 151, 197), (134, 9, 145), (253, 181, 88),
          (246, 11, 137), (55, 72, 220), (136, 8, 253), (56, 73, 180), (85, 241, 53),
          (153, 207, 15), (187, 183, 180), (149, 32, 71), (92, 113, 184), (131, 7, 201),
          (56, 20, 219), (243, 201, 77), (13, 74, 96), (79, 14, 44), (195, 150, 66),
          (2, 249, 42), (195, 135, 43), (105, 70, 66), (120, 107, 116), (122, 241, 22),
          (17, 19, 179), (162, 185, 124), (31, 65, 117), (88, 200, 80), (232, 49, 154),
          (72, 1, 46), (59, 144, 187), (200, 193, 118), (123, 165, 219), (194, 84, 34),
          (91, 184, 108), (252, 64, 153), (251, 121, 27), (105, 93, 210), (89, 85, 81),
          (58, 12, 154), (81, 3, 50), (200, 40, 236), (155, 147, 180), (73, 29, 176),
          (193, 19, 175), (157, 225, 121), (128, 195, 235), (146, 251, 108), (13, 146, 186),
          (231, 118, 145), (253, 15, 105), (187, 149, 62), (121, 247, 158), (34, 8, 142),
          (83, 61, 48), (119, 218, 69), (197, 94, 130), (222, 176, 142), (21, 20, 77),
          (6, 42, 17), (136, 33, 156), (39, 252, 211), (52, 50, 40), (183, 115, 34),
          (107, 80, 164), (195, 215, 74), (7, 154, 135), (136, 35, 24), (131, 241, 125),
          (208, 99, 208), (5, 4, 129), (137, 156, 175), (29, 141, 67), (44, 20, 99)]
default_aug_cfg = {
    'hsv_h': 0.014,
    'hsv_s': 0.68,
    'hsv_v': 0.36,
    'degree': (-10, 10),
    'translate': 0,
    'shear': 0.0,
    'beta': (8, 8),
}

rgb_mean = [0.485, 0.456, 0.406]
rgb_std = [0.229, 0.224, 0.225]
cv.setNumThreads(0)


# noinspection PyTypeChecker
class COCODataSets(Dataset):
    def __init__(self, img_root, annotation_path, img_list,
                 min_threshes=None,
                 max_thresh=1333,
                 augments=True,
                 use_crowd=True,
                 debug=False,
                 remove_blank=True,
                 aug_cfg=None,
                 ):
        super(COCODataSets, self).__init__()
        if isinstance(img_list, str):
            img_list = [img_list]
            
        self.img_name_list = list()
        for l in img_list:
            name_list = list()
            with open(l, 'r') as f:
                data = f.readlines()
                for line in data:
                    line = line.strip('\n')
                    name_list.append(line)
            self.img_name_list.append(name_list)
        
        self.coco = COCO(annotation_path)
        if min_threshes is None:
            min_threshes = [480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800]
        self.min_threshes = min_threshes
        self.max_thresh = max_thresh
        self.img_root = img_root
        self.use_crowd = use_crowd
        self.remove_blank = remove_blank
        self.augments = augments
        if aug_cfg is None:
            aug_cfg = default_aug_cfg
        self.aug_cfg = aug_cfg
        self.debug = debug
        self.empty_images_len = 0
        data_len = len(self.coco.imgs.keys())
        box_info_list = self.__load_data()
        self.box_info_list = box_info_list
        if len(box_info_list) != data_len:
            print("all data len:{:d} | valid data len:{:d}".format(data_len, len(box_info_list)))

        if self.debug:
            assert debug <= len(box_info_list), "not enough data to debug"
            print("debug")
            self.box_info_list = box_info_list[:debug]
        self.transform = None
        self.set_transform()

    def __load_data(self):
        box_info_list = list()
        for img_id in self.coco.imgs.keys():
            file_name = self.coco.imgs[img_id]['file_name']
            
            ctx_label = -1
            for i in range(len(self.img_name_list)):
                if file_name in self.img_name_list[i]:
                    ctx_label = i + 1
                    break
                    
            if ctx_label == -1:
                print("Loading data error !!!")
            
            width, height = self.coco.imgs[img_id]['width'], self.coco.imgs[img_id]['height']
            file_path = os.path.join(self.img_root, file_name)
            if not os.path.exists(file_path):
                print("img {:s} is not exist".format(file_path))
                continue

            assert width > 1 and height > 1, "invalid width or heights"
            anns = self.coco.imgToAnns[img_id]
            label_list = list()
            for ann in anns:
                category_id, box, iscrowd = ann['category_id'], ann['bbox'], ann['iscrowd']
                label_id = coco_ids.index(category_id)
                assert label_id >= 0, 'error label_id'
                if not self.use_crowd and iscrowd == 1:
                    continue
                x1, y1 = box[:2]
                x2, y2 = x1 + box[2], y1 + box[3]
                x1, x2 = min(x1, x2), max(x1, x2)
                y1, y2 = min(y1, y2), max(y1, y2)
                if x2 - x1 < 1 or y2 - y1 < 1:
                    print("not a valid box ", box)
                    continue
                if x1 < 0 or x2 > width or y1 < 0 or y2 > height:
                    print("warning box ", box)
                label_list.append((label_id, x1, y1, x2, y2))
            valid_box_len = len(label_list)
            if valid_box_len == 0:
                box_info = BoxInfo(img_path=file_path, boxes=np.zeros((0, 4)), labels=np.zeros((0,)), ctx_labels=np.zeros((0,)), weights=np.ones(shape=(0,)))
            else:
                label_info = np.array(label_list)
                ctx_label_info = np.array(len(label_info) * [ctx_label])
#                 print(ctx_label_info)
#                 print(label_info[:, 0])
                box_info = BoxInfo(img_path=file_path, boxes=label_info[:, 1:], labels=label_info[:, 0], ctx_labels=ctx_label_info, weights=np.ones_like(label_info[:, 0]))
            if self.remove_blank and valid_box_len == 0:
                self.empty_images_len += 1
                continue
            box_info_list.append(box_info)
        return box_info_list

    def __getitem__(self, item):
        box_info = self.box_info_list[item].clone().load_img()
        box_info = self.transform(box_info)
        return box_info

    @staticmethod
    def collect_fn(batch: List[BoxInfo]):
        batch_img = list()
        batch_target = list()
        batch_length = list()
        valid_size = list()
        max_h = max([item.img.shape[0] for item in batch])
        max_w = max([item.img.shape[1] for item in batch])
        for item in batch:
            padding_val = item.padding_val
            img = np.ones((max_h, max_w, 3)) * np.array(padding_val)
            ori_img = item.img
            ori_h, ori_w = ori_img.shape[:2]
            img[0:ori_h, 0:ori_w, :] = ori_img
            img = (img[:, :, ::-1] / 255.0 - np.array(rgb_mean)) / np.array(rgb_std)
            batch_img.append(img)
            
#             print(item.ctx_labels[:, None])
#             print(item.labels[:, None])
            target = np.concatenate([item.labels[:, None], item.boxes, item.ctx_labels[:, None]], axis=-1)
            batch_target.append(target)
            batch_length.append(len(target))
            valid_size.append((ori_w, ori_h))
            # valid_size.append((max_w, max_h))
        batch_img = torch.from_numpy(np.stack(batch_img, axis=0)).permute(0, 3, 1, 2).contiguous().float()
        batch_target = torch.from_numpy(np.concatenate(batch_target, axis=0)).float()
        return batch_img, valid_size, batch_target, batch_length

    def __len__(self):
        return len(self.box_info_list)

    def set_transform(self):
        color_gitter = OneOf(
            transforms=[
                Identity(),
                RandHSV(hgain=self.aug_cfg['hsv_h'],
                        vgain=self.aug_cfg['hsv_v'],
                        sgain=self.aug_cfg['hsv_s']),
                RandBlur().reset(p=0.5),
                RandNoise().reset(p=0.5)
            ]
        )
        basic_transform = Compose(
            transforms=[
                color_gitter,
                RandCrop(min_thresh=0.6, max_thresh=1.0).reset(p=0.2),
                RandScaleMinMax(min_threshes=self.min_threshes, max_thresh=self.max_thresh),
                RandPerspective(degree=self.aug_cfg['degree'], scale=(1.0, 1.0))
            ]
        )

        mosaic = MosaicWrapper(candidate_box_info=self.box_info_list,
                               sizes=[self.max_thresh],
                               color_gitter=color_gitter)

        augment_transform = Compose(
            transforms=[
                OneOf(transforms=[
                    (0.2, basic_transform),
                    (0.8, mosaic)
                ]),
                LRFlip().reset(p=0.5)
            ]
        )
        std_transform = RandScaleMinMax(min_threshes=[640], max_thresh=self.max_thresh)

        if self.augments:
            self.transform = augment_transform
        else:
            self.transform = std_transform
