import math
import torch
import numpy as np
from einops import rearrange, repeat
from torch import nn, einsum


# __all__ = ["OPS", "ReLUConvBN", "DilConv", "SepConv", "Identity", "Zero", "FactorizedReduce", "SwinBlock"]


OPS = {
    'none' : lambda C, stride, affine: Zero(stride),
    'skip_connect' : lambda C, stride, affine: (Identity() if stride == 1 
        else FactorizedReduce(C, C, affine=affine)),
    "attention_layer_sparse": lambda C, stride, affine: SwinBlock(
        C, C, 3, stride, 1, affine=affine, dim=96, depth=2, head=2
    ),
    "attention_layer_dense": lambda C, stride, affine: SwinBlock(
        C, C, 3, stride, 1, affine=affine, dim=192, depth=4, head=4
    ),
}


class SwinBlock(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, stride, padding, affine, dim, depth, head):
        super(SwinBlock, self).__init__()
        self.C_in = C_in
        self.C_out = C_out
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.affine = affine
        self.dim = dim
        self.depth = depth
        self.head = head

        self.conv_in = ReLUConvBN(
                C_in,
                dim,
                (1, 1),
                (1, 1),
                (0, 0),
                affine,
        )
        
        from nas_ood_single.search.models.swin import BasicLayer, PatchMerging
        self.swin_stage = BasicLayer(dim,
                                     depth=depth,
                                     num_heads=head,
                                     window_size=7,
                                     mlp_ratio=4.,
                                     qkv_bias=True,
                                     qk_scale=None,
                                     drop=0.,
                                     attn_drop=0.,
                                     drop_path=0.,
                                     norm_layer=nn.LayerNorm,
                                     downsample=PatchMerging,
                                     use_checkpoint=False)

        self.conv_out = ReLUConvBN(
            dim,
            C_out,
            kernel_size,
            stride,
            padding,
            affine,
        )

    def forward(self, x):
        x = self.conv_in(x)
        Wh, Ww = x.size(2), x.size(3)
        x = x.flatten(2).transpose(1, 2)
        x_out, H, W, x, Wh, Ww = self.swin_stage(x, Wh, Ww)
        x_out = x_out.view(-1, H, W, self.dim).permute(0, 3, 1, 2).contiguous()
        out = self.conv_out(x_out)
        return out


class SelfAttention(nn.Module):
    def __init__(
            self,
            C_in,
            C_out,
            kernel_size,
            stride,
            padding,
            affine,
            dim=512,
            attention_dropout=0.0
    ):
        super(SelfAttention, self).__init__()
        self.C_in = C_in
        self.C_out = C_out

        self.conv_w = nn.ModuleList(3 * [
            ReLUConvBN(
                C_in,
                C_in,
                (1, 1),
                (1, 1),
                (0, 0),
                affine,
            )
        ])

#         self.w = nn.ModuleList(3 * [nn.Linear(C_in, dim)])
        self.scaled_dot_product_attn = ScaledDotProductAttention(attention_dropout)

#         self.w_out = nn.Linear(dim, C_out)
        self.conv_out = ReLUConvBN(
                C_out,
                C_out,
                kernel_size,
                stride,
                padding,
                affine,
            )

    def forward(self, x):
        q = torch.reshape(self.conv_w[0](x), (x.shape[0], x.shape[1], x.shape[2] * x.shape[3]))
        k = torch.reshape(self.conv_w[1](x), (x.shape[0], x.shape[1], x.shape[2] * x.shape[3]))
        v = torch.reshape(self.conv_w[2](x), (x.shape[0], x.shape[1], x.shape[2] * x.shape[3]))

        out = torch.reshape(self.scaled_dot_product_attn(q, k, v), x.shape)
        out = self.conv_out(out)
        return out


class ScaledDotProductAttention(nn.Module):
    """Scaled dot-product attention mechanism."""

    def __init__(self, attention_dropout=0.0):
        super(ScaledDotProductAttention, self).__init__()
        self.dropout = nn.Dropout(attention_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, q, k, v):
        """

        Args:
        q: Queries, [B, L_q, D_q]
        k: Keys，[B, L_k, D_k]
        v: Values，[B, L_v, D_v]
        scale:

        Returns:
        """
        d_k = q.shape[-1]
        attention = torch.bmm(q, k.transpose(1, 2))
        # scale
        attention = attention / math.sqrt(d_k)
        # softmax
        attention = self.softmax(attention)
        # dropout
        attention = self.dropout(attention)
        # dot-product
        context = torch.bmm(attention, v)
        return context


class ReLUConvBN(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
        super(ReLUConvBN, self).__init__()
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=padding, bias=False),
            nn.BatchNorm2d(C_out, affine=affine))

    def forward(self, x):
        return self.op(x)


class DilConv(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, stride, padding, dilation, affine=True):
        super(DilConv, self).__init__()
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride,
                padding=padding, dilation=dilation, groups=C_in, bias=False),
            nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=affine))

    def forward(self, x):
        return self.op(x)


class SepConv(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
        super(SepConv, self).__init__()
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, groups=C_in, bias=False),
            nn.Conv2d(C_in, C_in, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_in, affine=affine),
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=1, padding=padding, groups=C_in, bias=False),
            nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=affine))

    def forward(self, x):
        return self.op(x)


class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class Zero(nn.Module):
    def __init__(self, stride):
        super(Zero, self).__init__()
        self.stride = stride

    def forward(self, x):
        if self.stride == 1:
            return x.mul(0.)
        return x[:,:,::self.stride,::self.stride].mul(0.)


class FactorizedReduce(nn.Module):

    def __init__(self, C_in, C_out, affine=True):
        super(FactorizedReduce, self).__init__()
        assert C_out % 2 == 0
        self.relu = nn.ReLU(inplace=False)
        self.conv_1 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        self.conv_2 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False) 
        self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x):
        x = self.relu(x)
        # out = torch.cat([self.conv_1(x), self.conv_2(x[:,:,1:,1:])], dim=1)
        ##############################
        out = torch.cat([self.conv_1(x), self.conv_2(x)], dim=1)
        out = self.bn(out)
        return out
