import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torch.distributed as dist

import moxing as mox

from .optimizer import Adam
from .models import NetWork

from .models import SinkhornDistance

from ..utils import *

from torch.autograd import Variable

import numpy as np

from coco_dataloader.coco_eval import object_detection_evaluate

import os
import json
from tempfile import NamedTemporaryFile
from pycocotools.coco import COCO


from torchvision.models.detection.transform import GeneralizedRCNNTransform
from torchvision.models.detection.image_list import ImageList

from keras.utils.np_utils import *


__all__ = ["InnerTrainer"]

class InnerTrainer:
    def __init__(self, cfg, device=None):
        self.cfg = cfg
        self.auxiliary = cfg.auxiliary
        self.auxiliary_weight = cfg.auxiliary_weight
        self.lambda_ctx = cfg.lambda_ctx
        self.lambda_gradorth = cfg.lambda_gradorth

        self.grad_clip = cfg.grad_clip
        self.report_freq = cfg.report_freq
        self.save_path = cfg.save_path
        self.max_epoch = cfg.epochs
        self.gpu_nums = cfg.gpus
        
        self.local_rank = dist.get_rank()
        
        
        model_without_ddp = NetWork(cfg.dataset, cfg.init_channels, cfg.num_classes, cfg.num_contexts, cfg.layers,
                                 proj_dims=cfg.proj_dims, steps=cfg.steps, multiplier=cfg.multiplier, min_size=cfg.min_size, max_size=cfg.max_size).cuda()

        if self.gpu_nums > 1:
            self.model = torch.nn.parallel.DistributedDataParallel(model_without_ddp, device_ids=[device], find_unused_parameters=True)

        self.start_epoch = cfg.start_epoch

        print("Param size = {}MB".format(count_parameters_in_MB(self.model)))

        weights = []
        a_weights = []
        for k, p in self.model.named_parameters():
            if ('alpha' not in k) and ('main' not in k) and ('input' not in k):
                weights.append(p)
            if 'alpha' in k:
                a_weights.append(p)
                
        self.w_optimizer = optim.SGD(
            weights, cfg.learning_rate,
            momentum=cfg.momentum, weight_decay=cfg.weight_decay
        )
        self.scheduler = optim.lr_scheduler.CosineAnnealingLR(
            self.w_optimizer, float(cfg.epochs), eta_min=cfg.learning_rate_min
        )

        self.alpha_optimizer = Adam(a_weights,
                                    lr=cfg.arch_learning_rate, betas=(0.5, 0.999), weight_decay=cfg.arch_weight_decay)

    def train_epoch(self, train_queue, epoch, all_freeze, writer, logging):
        losses = AverageMeter()

        lr = self.scheduler.get_lr()
        print('epoch: ', epoch, 'lr:', lr)

        self.model.train()
        global_count = 0

        for batch_id, (input, valid_size, target, batch_len) in enumerate(train_queue):
            input = input.to("cuda")
            target = target.to("cuda")

            global_count = global_count + 1

            if all_freeze == False:
                self.alpha_optimizer.zero_grad()
                self.w_optimizer.zero_grad()

                # without generator
                output = self.model(input, valid_size, {"target": target, "batch_len": batch_len}, all_freeze)
                penalty = sum([v for k, v in output.items() if "gradorth" in k])
                ctx_loss = sum([v for k, v in output.items() if "ctx" in k])
                box_loss = sum([v for k, v in output.items() if "box" in k])
                cls_loss = sum([v for k, v in output.items() if "cls" in k])

                loss = box_loss + cls_loss + self.lambda_ctx * ctx_loss + self.lambda_gradorth * penalty
                loss.backward()
                self.alpha_optimizer.step()
                nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip)
                self.w_optimizer.step()

            else:
                self.w_optimizer.zero_grad()
                output = self.model(input, valid_size, {"target": target, "batch_len": batch_len}, all_freeze)
                penalty = sum([v for k, v in output.items() if "gradorth" in k])
                ctx_loss = sum([v for k, v in output.items() if "ctx" in k])
                box_loss = sum([v for k, v in output.items() if "box" in k])
                cls_loss = sum([v for k, v in output.items() if "cls" in k])
                loss = box_loss + cls_loss + self.lambda_ctx * ctx_loss + self.lambda_gradorth * penalty
                loss.backward()
                nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip)
                self.w_optimizer.step()


            n = len(input)

            losses.update(loss.item(), n)

            if batch_id % self.report_freq == 0:
                logging.info("Train[{:0>3d}] Loss: {:.4f}".format(
                    batch_id, losses.avg
                ))

        if all_freeze == False:
            writer.add_scalar('data/stage1_loss', float(losses.avg), epoch)
        else:
            writer.add_scalar('data/stage2_loss', float(losses.avg), epoch)

        if all_freeze:
            self.scheduler.step()

        return losses.avg

    def validate(self, valid_queue, all_freeze, epoch, trlog, logging):
        n_threads = torch.get_num_threads()
        torch.set_num_threads(1)
        
        predict_list = list()
        target_list = list()
        
        self.model.eval()
        cpu_device = torch.device("cpu")

        
        with torch.no_grad():
            for batch_id, (img_tensor, valid_size, targets_tensor, batch_len) in enumerate(valid_queue):
                img_tensor = img_tensor.to("cuda")
                targets_tensor = targets_tensor.to("cuda")

                predicts = self.model(img_tensor, valid_size, {"target": targets_tensor, "batch_len": batch_len}, all_freeze)
                for pred, target in zip(predicts, targets_tensor.split(batch_len)):
                    predict_list.append(pred.cpu().numpy())
                    target_list.append(target.cpu().numpy())
            
        print("Synchronizing...")
        torch.cuda.synchronize(torch.device('cuda'))
        
        if self.local_rank == 0:
            coco_stats = object_detection_evaluate(target_list, predict_list)

            logging.info(
                " Valid AP: {:.4f} AP50: {:.4f} AP75: {:.4f} APs: {:.4f} APm: {:.4f} APl: {:.4f}".format(
                    coco_stats[0], coco_stats[1], coco_stats[2], coco_stats[3], coco_stats[4], coco_stats[5]
                ))

            trlog['ap'].append(coco_stats[0])
            trlog['ap50'].append(coco_stats[1])
            trlog['ap75'].append(coco_stats[2])
            trlog['aps'].append(coco_stats[3])
            trlog['apm'].append(coco_stats[4])
            trlog['apl'].append(coco_stats[5])
            
            branch_save = os.path.join(self.save_path, 'branch_weights')
            if os.path.exists(branch_save) == False:
                os.path.mkdir(branch_save)
                
            if epoch == 0 or epoch % 50 == 0:
                ctx_weight = dict()
                cls_weight = dict()
                
                for k, v in self.model.state_dict():
                    if 'cascade_head' and 'cls' in k:
                        cls_weight[k] = v
                    if 'cascade_head' and 'ctx' in k:
                        ctx_weight[k] = v
                        
                torch.save(ctx_weight, os.path.join(branch_save, 'ctx_' + str(epoch) + '.pth'))
                torch.save(cls_weight, os.path.join(branch_save, 'cls_' + str(epoch) + '.pth'))

            if coco_stats[5] >= trlog['max_apl']:
                trlog['max_ap'] = coco_stats[0]
                trlog['max_ap50'] = coco_stats[1]
                trlog['max_ap75'] = coco_stats[2]
                trlog['max_aps'] = coco_stats[3]
                trlog['max_apm'] = coco_stats[4]
                trlog['max_apl'] = coco_stats[5]
                trlog['max_ap_epoch'] = epoch
                
                name = 'max_apl'
                torch.save(self.model.state_dict(), os.path.join(self.save_path, name + '.pth'))
                
                name_normal = 'max_acc_normal'
                torch.save(self.model.alphas_normal, os.path.join(self.save_path, name_normal + '.pth'))

                name_reduce = 'max_acc_reduce'
                torch.save(self.model.alphas_reduce, os.path.join(self.save_path, name_reduce + '.pth'))
                

            if epoch >= (self.max_epoch - 10):
                if coco_stats[0] >= trlog['max_ap_last10']:
                    trlog['max_ap_last10'] = coco_stats[0]
                    trlog['max_ap_last10_epoch'] = epoch

            torch.save(trlog, os.path.join(self.save_path, 'trlog'))
            torch.set_num_threads(n_threads)
            
            return coco_stats
        else:
            return [0]
