import torch
import torch.nn as nn

from nas_ood_single.search.models.model_search import default_cfg
from ..operations import *
from ...utils import drop_path

from nas_ood_single.search.genotypes import PRIMITIVES
from nas_ood_single.search.models.cascade.cascade_rcnn import *
from nas_ood_single.search.models.swin import *


__all__ = ["NetworkEvaluate"]


class MixedOp(nn.Module):
    def __init__(self, C, stride, op_names, indices, weights):
        super(MixedOp, self).__init__()
        self._ops = nn.ModuleList()
        
        self.weight = list()
        for primitive in PRIMITIVES:
            if primitive in op_names:
                i = op_names.index(primitive)
                self.weight.append(weights[i])
            else:
                self.weight.append(0)
            op = OPS[primitive](C, stride, True)
            if 'pool' in primitive:
                op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
            if 'skip' in primitive and isinstance(op, Identity):
                op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
            self._ops.append(op)
        
#         for name, index in zip(op_names, indices):
#             print(name)
#             op = OPS[name](C, stride, True)
#             ## extra BN for non-param OPS ##
#             if 'pool' in name:
#                 op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
#             if 'skip' in name and isinstance(op, Identity):
#                 op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
#             self._ops += [op]
            
    def forward(self, x):
        feats = []
        for w, op in zip(self.weight, self._ops):
            if w == 0:
                continue
            feat = w * op(x)
            feats.append(feat)
        return sum(feats)


class Cell(nn.Module):
    def __init__(self, genotype, C_prev_prev, C_prev, C, reduction, reduction_prev):
        super(Cell, self).__init__()
        # print(C_prev_prev, C_prev, C)
        self.reduction = reduction

        if reduction_prev:
            self.preprocess0 = FactorizedReduce(C_prev_prev, C, affine=True)
        else:
            self.preprocess0 = ReLUConvBN(C_prev_prev, C, 1, 1, 0, affine=True)
        self.preprocess1 = ReLUConvBN(C_prev, C, 1, 1, 0, affine=True)

        if reduction:
            op_names, indices, weights = zip(*genotype.reduce)
            concat = genotype.reduce_concat
        else:
            op_names, indices, weights = zip(*genotype.normal)
            concat = genotype.normal_concat
        self._compile(C, op_names, indices, weights, concat, reduction)

    def _compile(self, C, op_names, indices, weights, concat, reduction):
        assert len(op_names) == len(indices)
#         self._steps = len(op_names) // 2
        self._steps = 4
        self._concat = concat
#         self.multiplier = len(concat)
        self._multiplier = 4

        self._ops = nn.ModuleList()
#         for name, index in zip(op_names, indices):
#             stride = 2 if reduction and index < 2 else 1
#             op = OPS[name](C, stride, True)
#             ## extra BN for non-param OPS ##
#             if 'pool' in name:
#                 op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
#             if 'skip' in name and isinstance(op, Identity):
#                 op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
#             ################################
#             self._ops += [op]
#         self._indices = indices
        for i in range(self._steps):
            for j in range(2 + i):
                stride = 2 if self.reduction and j < 2 else 1
                op = MixedOp(C, stride, op_names, indices, weights)
                self._ops.append(op)

    def forward(self, s0, s1, drop_prob):
        s0 = self.preprocess0(s0)
        s1 = self.preprocess1(s1)

        states = [s0, s1]
        offset = 0
        for i in range(self._steps):
#             h1 = states[self._indices[2*i]]
#             h2 = states[self._indices[2*i+1]]
#             op1 = self._ops[2*i]
#             op2 = self._ops[2*i+1]
#             h1 = op1(h1)
#             h2 = op2(h2)

#             s = h1 + h2
#             states += [s]
            s = sum(self._ops[offset + j](h) for j, h in enumerate(states))
            offset += len(states)
            states.append(s)
#         return torch.cat([states[i] for i in self._concat], dim=1)
        return torch.cat(states[-self._multiplier:], dim=1)


class NetworkEvaluate(nn.Module):
    def __init__(self, C, num_classes, layers, proj_dims, embed_dim, genotype):
        super(NetworkEvaluate, self).__init__()
        self._C = C
        self.num_cls = num_classes
        self._layers = layers
        self.proj_dims = proj_dims
        self.drop_path_prob = 0.0
        
        C_curr = embed_dim
        
        self.stem1 = nn.Sequential(
          nn.Conv2d(C_curr, C_curr, 3, stride=2, padding=1, bias=False),
          nn.BatchNorm2d(C_curr)
        )
        reduction_prev = True

        C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
        self.cells = nn.ModuleList()
        
        self.return_layers = [0, 1, 2, 3]
        self.in_channels_list = []
        for i in range(layers):
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            else:
                reduction = False

#             print(C_curr)
            cell = Cell(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
            
            if i in self.return_layers:
                self.in_channels_list.append(cell._multiplier * C_curr)
            
            reduction_prev = reduction
            self.cells.append(cell)
            C_prev_prev, C_prev = C_prev, cell._multiplier * C_curr

        self.cfg = {**default_cfg}

        # patch
        self.patch_embed = PatchEmbed(
            patch_size=4, in_chans=3, embed_dim=embed_dim,
            norm_layer=nn.LayerNorm)
        
        # fpn
        self.out_channels = 256
        self.fpn = FPN(in_channels=self.in_channels_list,
                       out_channel=self.out_channels,
                       bias=self.cfg['fpn_bias'])
        self.anchor_generator = AnchorGenerator(
            anchor_sizes=self.cfg['anchor_sizes'],
            anchor_scales=self.cfg['anchor_scales'],
            anchor_ratios=self.cfg['anchor_ratios'],
            strides=self.cfg['strides']
        )
        
        self.anchors = None
        rpn_head = RPNHead(self.cfg['fpn_channel'], anchor_per_grid=self.anchor_generator.anchor_per_grid)
        rpn_pre_nms_top_n = {"train": self.cfg['rpn']['pre_nms_top_n_train'],
                             "test": self.cfg['rpn']['pre_nms_top_n_test']}
        rpn_post_nms_top_n = {"train": self.cfg['rpn']['post_nms_top_n_train'],
                              "test": self.cfg['rpn']['post_nms_top_n_test']}
        self.feature_keys = ['0', '1', '2', '3', 'pool']

        self.rpn = RPN(
            rpn_head=rpn_head,
            pre_nms_top_n=rpn_pre_nms_top_n,
            post_nms_top_n=rpn_post_nms_top_n,
            pos_iou_thr=self.cfg['rpn']['pos_iou_thr'],
            neg_iou_thr=self.cfg['rpn']['neg_iou_thr'],
            allow_low_quality_matches=self.cfg['rpn']['allow_low_quality_matches'],
            nms_thresh=self.cfg['rpn']['nms_thresh'],
            sample_size=self.cfg['rpn']['sample_size'],
            positive_fraction=self.cfg['rpn']['positive_fraction'],
            bbox_weights=self.cfg['rpn']['bbox_weights'],
            iou_type=self.cfg['rpn']['iou_type']
        )

        roi = MultiScaleRoIAlign(
            featmap_names=self.cfg['roi']['featmap_names'],
            output_size=self.cfg['roi']['output_size'],
            sampling_ratio=self.cfg['roi']['sampling_ratio']
        )
        roi.channel_size = self.cfg['fpn_channel']
        self.cascade_head = CascadeHead(self.num_cls,
                                        3,
                                        roi,
                                        self.cfg['cascade_head']['stage_loss_weights'],
                                        num_stages=self.cfg['cascade_head']['num_stages'],
                                        fc_out_channels=self.cfg['cascade_head']['fc_out_channels'],
                                        detail_cfg=self.cfg['cascade_head']['bbox_head'])
        
    def post_process(self, cls_predicts, box_predicts, valid_size):
        predicts = list()
        for cls, box, wh in zip(cls_predicts, box_predicts, valid_size):
            box[..., [0, 2]] = box[..., [0, 2]].clamp(min=0, max=wh[0])
            box[..., [1, 3]] = box[..., [1, 3]].clamp(min=0, max=wh[1])
            scores = cls[:, 1:]
            labels = torch.arange(scores.shape[-1], device=cls.device)
            labels = labels.view(1, -1).expand_as(scores)
            boxes = box.unsqueeze(1).repeat(1, scores.shape[-1], 1).reshape(-1, 4)
            scores = scores.reshape(-1)
            labels = labels.reshape(-1)
            inds = torch.nonzero(scores > self.cfg['box_score_thresh'], as_tuple=False).squeeze(1)
            boxes, scores, labels = boxes[inds], scores[inds], labels[inds]
            keep = ((boxes[..., 2] - boxes[..., 0]) > 1e-2) & ((boxes[..., 3] - boxes[..., 1]) > 1e-2)
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]
            from torchvision.ops.boxes import batched_nms
            keep = batched_nms(boxes, scores, labels, self.cfg['box_nms_thresh'])
            keep = keep[:self.cfg['box_detections_per_img']]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]
            pred = torch.cat([boxes, scores[:, None], labels[:, None]], dim=-1)
            predicts.append(pred)
        return predicts

    def forward(self, input, valid_size):
#         print(input)
        if self.anchors:
            anchor_num = sum([a.shape[0] for a in self.anchors]) / self.anchor_generator.anchor_per_grid
        else:
            anchor_num = -1  
        
        s0 = self.patch_embed(input)
        s1 = self.stem1(s0)

        layer_order = 0
        features = []
        for i, cell in enumerate(self.cells):
            s0, s1 = s1, cell(s0, s1, self.drop_path_prob)

            if i in self.return_layers:
                features.append(s1)
                layer_order += 1   
            if layer_order == len(self.return_layers):
                break
                
        features = self.fpn(features)
#         print(features)
        
        features_resolution = sum([i.shape[-2] * i.shape[-1] for i in features])
        if features_resolution != anchor_num:
            self.anchors = self.anchor_generator(features)

        boxes, _ = self.rpn(features, self.anchors, valid_size, None)
        
#         print(boxes)
        feature_dict = dict()
        for k, v in zip(self.feature_keys, features):
            feature_dict[k] = v

        box_predicts, cls_predicts, __ = self.cascade_head(feature_dict, boxes, valid_size, None)
        
#         print(box_predicts)
#         print(cls_predicts)

        return self.post_process(cls_predicts, box_predicts, valid_size)
