import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

from nas_ood_single.search.operations import *
from nas_ood_single.search.genotypes import Genotype, PRIMITIVES
from nas_ood_single.utils import drop_path

from nas_ood_single.search.models.classConcept import ccNet

from nas_ood_single.search.models.classCategory import cNet

from timm.models.layers import DropPath, to_2tuple, trunc_normal_

from nas_ood_single.search.models.cascade.cascade_rcnn import *
from nas_ood_single.search.models.swin import *
from torchvision.models.detection.transform import GeneralizedRCNNTransform

from torch.autograd import Function

from keras.utils.np_utils import *
import os



from typing import Tuple, List, Dict, Optional, Union
from collections import OrderedDict
import warnings

__all__ = ["MixedOp", "Cell", "NetWork"]


default_cfg = {
    "num_cls": 20,
    "num_ctx": 3,
    "backbone": "resnet18",
    "pretrained": False,
    "reduction": False,
    "norm_layer": None,
    "fpn_channel": 256,
    "fpn_bias": True,
    "anchor_sizes": [32., 64., 128., 256., 512.],
    "anchor_scales": [2 ** 0, ],
    "anchor_ratios": [0.5, 1.0, 2.0],
    "strides": [4., 8., 16., 32., 64.],
    "rpn": {
        "pre_nms_top_n_train": 2000,
        "post_nms_top_n_train": 2000,
        "pre_nms_top_n_test": 1000,
        "post_nms_top_n_test": 1000,
        "pos_iou_thr": 0.7,
        "neg_iou_thr": 0.3,
        "allow_low_quality_matches": True,
        "nms_thresh": 0.7,
        "sample_size": 256,
        "positive_fraction": 0.5,
        "bbox_weights": [1.0, 1.0, 1.0, 1.0],
        "iou_type": "giou"
    },
    "roi": {
        "featmap_names": ['0', '1', '2', '3'],
        "output_size": 7,
        "sampling_ratio": 2
    },
    "cascade_head": {
        "num_stages": 3,
        # "num_stages": 1,
        "stage_loss_weights": [1, 0.5, 0.25],
        # "stage_loss_weights": [1.0, ],
        "fc_out_channels": 1024,
        "bbox_head": [
            {
                "pos_iou_thr": 0.5,
                "neg_iou_thr": 0.5,
                "allow_low_quality_matches": False,
                "add_gt_as_proposals": True,
                "sample_size": 512,
                "positive_fraction": 0.25,
                "bbox_weights": [0.1, 0.1, 0.2, 0.2],
                "iou_type": "giou"
            },
            {
                "pos_iou_thr": 0.6,
                "neg_iou_thr": 0.6,
                "allow_low_quality_matches": False,
                "add_gt_as_proposals": True,
                "sample_size": 512,
                "positive_fraction": 0.25,
                "bbox_weights": [0.05, 0.05, 0.1, 0.1],
                "iou_type": "giou"
            },
            {
                "pos_iou_thr": 0.7,
                "neg_iou_thr": 0.7,
                "allow_low_quality_matches": False,
                "add_gt_as_proposals": True,
                "sample_size": 512,
                "positive_fraction": 0.25,
                "bbox_weights": [0.033, 0.033, 0.067, 0.067],
                "iou_type": "giou"
            },
        ]
    },
    "box_score_thresh": 0.05,
    "box_nms_thresh": 0.5,
    "box_detections_per_img": 100

}


class MixedOp(nn.Module):

    def __init__(self, C, stride):
        super(MixedOp, self).__init__()
        self._ops = nn.ModuleList()
        for primitive in PRIMITIVES:
            op = OPS[primitive](C, stride, True)
            if 'pool' in primitive:
                op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
            if 'skip' in primitive and isinstance(op, Identity):
                op = nn.Sequential(op, nn.BatchNorm2d(C, affine=True))
            self._ops.append(op)
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
                m.weight.requires_grad = False
                m.bias.requires_grad = False

    def forward(self, x, weights, drop_prob):
        if weights.sum() == 0:
            return 0
        feats = []
        for w, op in zip(weights, self._ops):
            if w == 0:
                continue
            feat = w * op(x)
            if self.training and drop_prob > 0:
                feat = drop_path(feat, drop_prob)
            feats.append(feat)
        return sum(feats)


class Cell(nn.Module):

    def __init__(self, steps, multiplier, C_prev_prev, C_prev, C, reduction, reduction_prev):
        super(Cell, self).__init__()
        self.reduction = reduction

        if reduction_prev:
            self.preprocess0 = FactorizedReduce(C_prev_prev, C, affine=True)
        else:
            self.preprocess0 = ReLUConvBN(C_prev_prev, C, 1, 1, 0, affine=True)
        self.preprocess1 = ReLUConvBN(C_prev, C, 1, 1, 0, affine=True)
        self._steps = steps
        self._multiplier = multiplier

        self._ops = nn.ModuleList()
        for i in range(self._steps):
            for j in range(2 + i):
                stride = 2 if reduction and j < 2 else 1
                op = MixedOp(C, stride)
                self._ops.append(op)

    def forward(self, s0, s1, weights, drop_prob):
        s0 = self.preprocess0(s0)
        s1 = self.preprocess1(s1)

        states = [s0, s1]
        offset = 0
        for i in range(self._steps):  # 4; weights 0-1, 2-4, 5-8, 9-13
            s = sum(self._ops[offset + j](h, weights[offset + j], drop_prob) for j, h in enumerate(states))
            offset += len(states)
            states.append(s)

        return torch.cat(states[-self._multiplier:], dim=1)


class ResidualBlock(nn.Module):
    def __init__(self, dim_in, dim_out):
        super(ResidualBlock, self).__init__()
        self.main = nn.Sequential(
            nn.Conv2d(dim_in, dim_out, kernel_size=3, stride=1, padding=1, bias=False),
            nn.InstanceNorm2d(dim_out, affine=True, track_running_stats=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(dim_out, dim_out, kernel_size=3, stride=1, padding=1, bias=False),
            nn.InstanceNorm2d(dim_out, affine=True, track_running_stats=True))

    def forward(self, x):
        return x + self.main(x)


class GradReverse(Function):

    @staticmethod
    def forward(ctx, x, lambd, **kwargs: None):
        ctx.lambd = lambd
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.clone()
        return -1.0 * grad_output, None

    
class NetWork(nn.Module):
    def __init__(self, dataset, C, num_classes, num_contexts, layers, proj_dims=2, steps=4, multiplier=4,
                 stem_multiplier=3, min_size=800, max_size=1333, img_size=224, patch_size=4, in_chans=3, embed_dim=96,
                 patch_norm=True, ape=True):
        super(NetWork, self).__init__()
        self._C = C
        self.num_cls = num_classes
        self.num_ctx = num_contexts
        self._layers = layers
        self._steps = steps
        self._multiplier = multiplier
        self.num_edges = sum(1 for i in range(self._steps) for n in range(2 + i))
        self.proj_dims = proj_dims
        self.dataset = dataset

        C_curr = embed_dim
        self.stem1 = nn.Sequential(
                     nn.Conv2d(C_curr, C_curr, kernel_size=3, stride=2, padding=1, bias=False),
                     nn.BatchNorm2d(C_curr),
                )
        reduction_prev = True

        C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
        self.cells = nn.ModuleList()

        self.return_layers = [0, 1, 2, 3]
        self.in_channels_list = []
        for i in range(layers):
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            else:
                reduction = False

            if i in self.return_layers:
                self.in_channels_list.append(multiplier * C_curr)

            cell = Cell(steps, multiplier, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
            reduction_prev = reduction
            self.cells.append(cell)
            C_prev_prev, C_prev = C_prev, multiplier * C_curr

        self._initialize_alphas()

        self.cfg = {**default_cfg}
        
        # patch
        self.patch_embed = PatchEmbed(
            patch_size=4, in_chans=3, embed_dim=embed_dim,
            norm_layer=nn.LayerNorm)
        
        # fpn
        self.out_channels = 256
        self.fpn = FPN(in_channels=self.in_channels_list,
                       out_channel=self.out_channels,
                       bias=self.cfg['fpn_bias'])
        self.anchor_generator = AnchorGenerator(
            anchor_sizes=self.cfg['anchor_sizes'],
            anchor_scales=self.cfg['anchor_scales'],
            anchor_ratios=self.cfg['anchor_ratios'],
            strides=self.cfg['strides']
        )
        
        self.anchors = None
        rpn_head = RPNHead(self.cfg['fpn_channel'], anchor_per_grid=self.anchor_generator.anchor_per_grid)
        rpn_pre_nms_top_n = {"train": self.cfg['rpn']['pre_nms_top_n_train'],
                             "test": self.cfg['rpn']['pre_nms_top_n_test']}
        rpn_post_nms_top_n = {"train": self.cfg['rpn']['post_nms_top_n_train'],
                              "test": self.cfg['rpn']['post_nms_top_n_test']}
        self.feature_keys = ['0', '1', '2', '3', 'pool']

        self.rpn = RPN(
            rpn_head=rpn_head,
            pre_nms_top_n=rpn_pre_nms_top_n,
            post_nms_top_n=rpn_post_nms_top_n,
            pos_iou_thr=self.cfg['rpn']['pos_iou_thr'],
            neg_iou_thr=self.cfg['rpn']['neg_iou_thr'],
            allow_low_quality_matches=self.cfg['rpn']['allow_low_quality_matches'],
            nms_thresh=self.cfg['rpn']['nms_thresh'],
            sample_size=self.cfg['rpn']['sample_size'],
            positive_fraction=self.cfg['rpn']['positive_fraction'],
            bbox_weights=self.cfg['rpn']['bbox_weights'],
            iou_type=self.cfg['rpn']['iou_type']
        )

        roi = MultiScaleRoIAlign(
            featmap_names=self.cfg['roi']['featmap_names'],
            output_size=self.cfg['roi']['output_size'],
            sampling_ratio=self.cfg['roi']['sampling_ratio']
        )
        roi.channel_size = self.cfg['fpn_channel']
        self.cascade_head = CascadeHead(self.num_cls,
                                        self.num_ctx,
                                        roi,
                                        self.cfg['cascade_head']['stage_loss_weights'],
                                        num_stages=self.cfg['cascade_head']['num_stages'],
                                        fc_out_channels=self.cfg['cascade_head']['fc_out_channels'],
                                        detail_cfg=self.cfg['cascade_head']['bbox_head'])        

    def new(self):
        model_new = NetWork(self._C, self._num_classes, self._layers).cuda()
        for x, y in zip(model_new.arch_parameters(), self.arch_parameters()):
            x.data.copy_(y.data)
        return model_new

    def forward(self, input, valid_size, targets, all_freeze, grad_reverse=False):
        if self.anchors:
            anchor_num = sum([a.shape[0] for a in self.anchors]) / self.anchor_generator.anchor_per_grid
        else:
            anchor_num = -1  
        
        s0 = self.patch_embed(input)
        s1 = self.stem1(s0)
        
        
        if not all_freeze:
            self.proj_alphas(self.A_normals, self.A_reduces)

        layer_order = 0
        features = []
        for i, cell in enumerate(self.cells):
            if cell.reduction:
                weights = self.alphas_reduce
            else:
                weights = self.alphas_normal

            s0, s1 = s1, cell(s0, s1, weights, self.drop_path_prob)

            if i in self.return_layers:
                features.append(s1)
                layer_order += 1   
            if layer_order == len(self.return_layers):
                break
                
        features = self.fpn(features)
        
        features_resolution = sum([i.shape[-2] * i.shape[-1] for i in features])
        if features_resolution != anchor_num:
            self.anchors = self.anchor_generator(features)

        boxes, rpn_losses = self.rpn(features, self.anchors, valid_size, targets)
        feature_dict = dict()
        for k, v in zip(self.feature_keys, features):
            feature_dict[k] = v

        box_predicts, cls_predicts, roi_losses = self.cascade_head(feature_dict, boxes, valid_size, targets)

        losses = {}
        losses.update(rpn_losses)
        losses.update(roi_losses)

        if self.training:
            return losses
        else:
            return self.post_process(cls_predicts, box_predicts, valid_size)

    def _initialize_alphas(self):
        self.alphas_normal_ = nn.ParameterList([nn.Parameter(1e-3 * torch.randn(self.proj_dims, device='cuda')) for i in range(self._steps)])
        self.alphas_reduce_ = nn.ParameterList([nn.Parameter(1e-3 * torch.randn(self.proj_dims, device='cuda')) for i in range(self._steps)])

    def freeze_alpha(self, normal_freeze_alpha, reduce_freeze_alpha):
        offset = 0
        for i, (flag, alpha) in enumerate(zip(normal_freeze_alpha, self.alphas_normal_)):
            if flag and alpha.requires_grad:
                alpha.requires_grad = False
                for cell in self.cells:
                    if cell.reduction:
                        continue
                    for j in range(offset, offset + i + 2):
                        op = cell._ops[j]
                        for m in op.modules():
                            if isinstance(m, nn.BatchNorm2d):
                                m.weight.requires_grad = True
                                m.bias.requires_grad = True

            offset += i + 2

        offset = 0
        for i, (flag, alpha) in enumerate(zip(reduce_freeze_alpha, self.alphas_reduce_)):
            if flag and alpha.requires_grad:
                alpha.requires_grad = False
                for cell in self.cells:
                    if not cell.reduction:
                        continue
                    for j in range(offset, offset + i + 2):
                        op = cell._ops[j]
                        for m in op.modules():
                            if isinstance(m, nn.BatchNorm2d):
                                m.weight.requires_grad = True
                                m.bias.requires_grad = True

            offset += i + 2

    def init_proj_mat(self, A_normals, A_reduces):
        self.A_normals = A_normals
        self.A_reduces = A_reduces

    def init_bias(self, normal_bias, reduce_bias):
        self.normal_bias = normal_bias
        self.reduce_bias = reduce_bias

    def proj_alphas(self, A_normals, A_reduces):
        assert len(A_normals) == len(A_reduces) == self._steps
        alphas_normal = []
        alphas_reduce = []
        alphas_normal_ = torch.stack(list(i for i in self.alphas_normal_))
        alphas_reduce_ = torch.stack(list(i for i in self.alphas_reduce_))
        for i in range(self._steps):
            A_normal = A_normals[i].to("cuda").requires_grad_(False)
            t_alpha = alphas_normal_[[i]].mm(A_normal).reshape(-1, len(PRIMITIVES))
            alphas_normal.append(t_alpha)
            A_reduce = A_reduces[i].to("cuda").requires_grad_(False)
            t_alpha = alphas_reduce_[[i]].mm(A_reduce).reshape(-1, len(PRIMITIVES))
            alphas_reduce.append(t_alpha)

        self.alphas_normal = torch.cat(alphas_normal) - self.normal_bias.to(
            alphas_normal_.device)
        self.alphas_reduce = torch.cat(alphas_reduce) - self.reduce_bias.to(
            alphas_reduce_.device)

    def genotype(self):
        def _parse(weights):
            gene = []
            n = 2
            start = 0
            for i in range(self._steps):
                end = start + n
                W = weights[start:end].copy()
                edges = sorted(range(i + 2),
                               key=lambda x: -max(W[x][k] for k in range(len(W[x]))
                                                  if k != PRIMITIVES.index('none')))[:2]
                for j in edges:
                    k_best = None
                    for k in range(len(W[j])):
                        if k != PRIMITIVES.index('none'):
                            if k_best is None or W[j][k] > W[j][k_best]:
                                k_best = k
                    gene.append((PRIMITIVES[k_best], j))
                start = end
                n += 1
            return gene

        gene_normal = _parse(F.softmax(self.alphas_normal, dim=-1).data.cpu().numpy())
        gene_reduce = _parse(F.softmax(self.alphas_reduce, dim=-1).data.cpu().numpy())

        concat = range(2 + self._steps - self._multiplier, self._steps + 2)
        genotype = Genotype(
            normal=gene_normal, normal_concat=concat,
            reduce=gene_reduce, reduce_concat=concat
        )
        return genotype

    @torch.jit.unused
    def eager_outputs(self, losses, detections):
        # type: (Dict[str, Tensor], List[Dict[str, Tensor]]) -> Union[Dict[str, Tensor], List[Dict[str, Tensor]]]
        if self.training:
            return losses

        return detections
    
    def post_process(self, cls_predicts, box_predicts, valid_size):
        predicts = list()
        for cls, box, wh in zip(cls_predicts, box_predicts, valid_size):
            box[..., [0, 2]] = box[..., [0, 2]].clamp(min=0, max=wh[0])
            box[..., [1, 3]] = box[..., [1, 3]].clamp(min=0, max=wh[1])
            scores = cls[:, 1:]
            labels = torch.arange(scores.shape[-1], device=cls.device)
            labels = labels.view(1, -1).expand_as(scores)
            boxes = box.unsqueeze(1).repeat(1, scores.shape[-1], 1).reshape(-1, 4)
            scores = scores.reshape(-1)
            labels = labels.reshape(-1)
            inds = torch.nonzero(scores > self.cfg['box_score_thresh'], as_tuple=False).squeeze(1)
            boxes, scores, labels = boxes[inds], scores[inds], labels[inds]
            keep = ((boxes[..., 2] - boxes[..., 0]) > 1e-2) & ((boxes[..., 3] - boxes[..., 1]) > 1e-2)
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]
            from torchvision.ops.boxes import batched_nms
            keep = batched_nms(boxes, scores, labels, self.cfg['box_nms_thresh'])
            keep = keep[:self.cfg['box_detections_per_img']]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]
            pred = torch.cat([boxes, scores[:, None], labels[:, None]], dim=-1)
            predicts.append(pred)
        return predicts

