import os
os.environ['MOSEKLM_LICENSE_FILE'] = 'mosek/mosek.lic'

import sys
import argparse
import os.path as osp
import torch.backends.cudnn as cudnn
import torchvision.datasets as datasets
from torch.utils.data import DataLoader
import torchvision.transforms as T
from torch.utils.data.distributed import DistributedSampler
import numpy as np
import torch
from tensorboardX import SummaryWriter

sys.path.insert(0, osp.join(osp.dirname(osp.abspath(__file__)), '../'))

from nas_ood_single import Trainer
from coco_dataloader.coco import COCODataSets

import datetime
import logging


def get_args():
    parser = argparse.ArgumentParser("NAS-DO")
    
    # multi-gpu
    parser.add_argument('--distributed', default=True, help='distributed training')
    parser.add_argument('--gpus', type=int, default=2, help='number of gpus')
    parser.add_argument('--local_rank', type=int, default=-1)
    
    # hyper-parameters
    parser.add_argument('--batch_size', type=int, default=1, help='samples per gpu')
    parser.add_argument('--proj_dims', type=int, default=3, help='proj dimensions, number of operations')
    parser.add_argument('--sparseness', type=int, default=2, help='sparseness')
    parser.add_argument('--num_classes', type=int, default=11, help='num_classes')
    parser.add_argument('--num_contexts', type=int, default=3, help='num_contexts')
    parser.add_argument('--learning_rate', type=float, default=0.001, help='init learning rate')
    parser.add_argument('--learning_rate_min', type=float, default=0.0001, help='min learning rate')
    parser.add_argument('--momentum', type=float, default=0.9, help='momentum')
    parser.add_argument('--weight_decay', type=float, default=3e-4, help='weight decay')
    parser.add_argument('--report_freq', type=float, default=50, help='report frequency')
    parser.add_argument('--epochs', type=int, default=600, help='num of training epochs')
    parser.add_argument('--init_channels', type=int, default=36, help='num of init channels')
    parser.add_argument('--layers', type=int, default=4, help='total number of layers')
    parser.add_argument('--steps', type=int, default=4, help='total number of layers')
    parser.add_argument('--multiplier', type=int, default=4)
    parser.add_argument('--drop_path_prob', type=float, default=0.2, help='drop path probability')
    parser.add_argument('--seed', type=int, default=5, help='random seed')
    parser.add_argument('--grad_clip', type=float, default=5, help='gradient clipping')
    parser.add_argument('--arch_learning_rate', type=float, default=3e-4, help='learning rate for arch encoding')
    parser.add_argument('--arch_weight_decay', type=float, default=1e-3, help='weight decay for arch encoding')   
#     parser.add_argument('--min_size', type=int, default=224)
#     parser.add_argument('--max_size', type=int, default=224)
    parser.add_argument('--ood_env', type=str, default='weather', help='OOD environment')
    parser.add_argument('--lambda_ctx', type=float, default=0.5)
    parser.add_argument('--lambda_gradorth', type=float, default=0.5)

    # file paths
    parser.add_argument('--save_path', type=str, default='save_dir/bdd100k/weather')
    parser.add_argument('--data_root_path', type=str, default='/home/ma-user/work/wufan/dataset/bdd100k')
    parser.add_argument('--image_root', type=str, default='/home/ma-user/work/wufan/dataset/bdd100k/images')
    parser.add_argument('--train_annos', type=str,
                        default='/home/ma-user/work/wufan/dataset/bdd100k/detection_coco_weather_train_sample.json', 
                        help='training annotation file')
    parser.add_argument('--test_annos', type=str,
                        default='/home/ma-user/work/wufan/dataset/bdd100k/detection_coco_weather_val_sample.json', 
                        help='testing annotation file')

    args, unknown_args = parser.parse_known_args()
    return args


def get_loader(args):
    if args.domain == 'weather':
        train_list = ['clear.txt', 'overcast.txt']
        test_list = ['foggy.txt', 'partly_cloudy.txt', 'rainy.txt', 'snowy.txt']
        train_list = [os.path.join(args.data_root_path+'/'+args.domain, i) for i in train_list]
        test_list = [os.path.join(args.data_root_path+'/'+args.domain, i) for i in test_list]
    elif args.domain == 'time-of-day':
        train_list = ['daytime.txt', 'dawn_dusk.txt']
        test_list = ['night.txt']
        train_list = [os.path.join('/cache/NAS-OoD-od-cascade-ci/dataset/bdd100k/timeofday', i) for i in train_list]
        test_list = [os.path.join('/cache/NAS-OoD-od-cascade-ci/dataset/bdd100k/timeofday', i) for i in test_list]
    else:
        raise ValueError("Domain is not supported !!")
        
    trainset = COCODataSets(args.image_root, args.train_annos, train_list, augments=True)
    testset = COCODataSets(args.image_root, args.test_annos, test_list, augments=False)

    if args.distributed:
        data_loader = DataLoader(dataset=trainset, batch_size=args.batch_size,
                                  num_workers=2, collate_fn=trainset.collect_fn, sampler=DistributedSampler(trainset))
        data_loader_test = DataLoader(dataset=testset, batch_size=args.batch_size,
                                 num_workers=2, collate_fn=testset.collect_fn, sampler=DistributedSampler(testset))
    else:
        train_sampler = torch.utils.data.RandomSampler(trainset)
        test_sampler = torch.utils.data.SequentialSampler(testset)
        train_batch_sampler = torch.utils.data.BatchSampler(train_sampler, args.batch_size, drop_last=True)
        
        data_loader = torch.utils.data.DataLoader(
            trainset, batch_sampler=train_batch_sampler, num_workers=2,
            collate_fn=trainset.collate_fn)

        data_loader_test = torch.utils.data.DataLoader(
            testset, batch_size=args.batch_size,
            sampler=test_sampler, num_workers=2,
            collate_fn=testset.collate_fn)

    return data_loader, data_loader_test


def main():
    args = get_args()

    log_format = '%(asctime)s %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                        format=log_format, datefmt='%m/%d %I:%M:%S %p')
    fh = logging.FileHandler(os.path.join(args.save_path, 'log.txt'))
    fh.setFormatter(logging.Formatter(log_format))
    logging.getLogger().addHandler(fh)

    logging.info(vars(args))

    if torch.cuda.is_available():
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True
        
    if args.distributed:
        torch.distributed.init_process_group(backend="nccl")
        local_rank = torch.distributed.get_rank()
        torch.cuda.set_device(local_rank)
        device = torch.device("cuda", local_rank)
    else:
        device = None


    # torch.cuda.set_device(args.gpu)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed(args.seed)

    writer = SummaryWriter(args.save_path)

    train_loader, test_loader = get_loader(args)

    trainer = Trainer(args, logging, writer, device)
    trainer.train(train_loader, test_loader, args.local_rank)


if __name__ == "__main__":
    main()
